@@ -8,18 +8,23 @@ UDP_LIMIT = 30
 DUMMY = lambda s: s
 
 asyncio.StreamReader.read_ = lambda self: self.read(PACKET_SIZE)
+asyncio.StreamReader.read_w = lambda self, n: asyncio.wait_for(self.read(n), timeout=SOCKET_TIMEOUT)
 asyncio.StreamReader.read_n = lambda self, n: asyncio.wait_for(self.readexactly(n), timeout=SOCKET_TIMEOUT)
 asyncio.StreamReader.read_until = lambda self, s: asyncio.wait_for(self.readuntil(s), timeout=SOCKET_TIMEOUT)
+asyncio.StreamReader.rollback = lambda self, s: self._buffer.__setitem__(slice(0, 0), s)
 
 class AuthTable(object):
     _auth = {}
+    _user = {}
     def __init__(self, remote_ip, authtime):
         self.remote_ip = remote_ip
         self.authtime = authtime
     def authed(self):
-        return time.time() - self._auth.get(self.remote_ip, 0) <= self.authtime
-    def set_authed(self):
+        if time.time() - self._auth.get(self.remote_ip, 0) <= self.authtime:
+            return self._user[self.remote_ip]
+    def set_authed(self, user):
         self._auth[self.remote_ip] = time.time()
+        self._user[self.remote_ip] = user
 
 async def prepare_ciphers(cipher, reader, writer, bind=None, server_side=True):
     if cipher:
@@ -51,18 +56,19 @@ def schedule(rserver, salgorithm, host_name, port):
     else:
         raise Exception('Unknown scheduling algorithm') #Unreachable
 
-async def stream_handler(reader, writer, unix, lbind, protos, rserver, cipher, sslserver, debug=0, authtime=86400*30, block=None, salgorithm='fa', verbose=DUMMY, modstat=lambda r,h:lambda i:DUMMY, **kwargs):
+async def stream_handler(reader, writer, unix, lbind, protos, rserver, cipher, sslserver, debug=0, authtime=86400*30, block=None, salgorithm='fa', verbose=DUMMY, modstat=lambda u,r,h:lambda i:DUMMY, **kwargs):
     try:
         reader, writer = proto.sslwrap(reader, writer, sslserver, True, None, verbose)
         if unix:
             remote_ip, server_ip, remote_text = 'local', None, 'unix_local'
         else:
-            remote_ip, remote_port, *_ = writer.get_extra_info('peername')
+            peername = writer.get_extra_info('peername')
+            remote_ip, remote_port, *_ = peername if peername else ('unknow_remote_ip','unknow_remote_port')
             server_ip = writer.get_extra_info('sockname')[0]
             remote_text = f'{remote_ip}:{remote_port}'
         local_addr = None if server_ip in ('127.0.0.1', '::1', None) else (server_ip, 0)
         reader_cipher, _ = await prepare_ciphers(cipher, reader, writer, server_side=False)
-        lproto, host_name, port, lbuf, rbuf = await proto.parse(protos, reader=reader, writer=writer, authtable=AuthTable(remote_ip, authtime), reader_cipher=reader_cipher, sock=writer.get_extra_info('socket'), **kwargs)
+        lproto, user, host_name, port, lbuf, rbuf = await proto.accept(protos, reader=reader, writer=writer, authtable=AuthTable(remote_ip, authtime), reader_cipher=reader_cipher, sock=writer.get_extra_info('socket'), **kwargs)
         if host_name == 'echo':
             asyncio.ensure_future(lproto.channel(reader, writer, DUMMY, DUMMY))
         elif host_name == 'empty':
@@ -83,7 +89,7 @@ async def stream_handler(reader, writer, unix, lbind, protos, rserver, cipher, s
             except Exception:
                 writer_remote.close()
                 raise Exception('Unknown remote protocol')
-            m = modstat(remote_ip, host_name)
+            m = modstat(user, remote_ip, host_name)
             lchannel = lproto.http_channel if rbuf else lproto.channel
             asyncio.ensure_future(lproto.channel(reader_remote, writer, m(2+roption.direct), m(4+roption.direct)))
             asyncio.ensure_future(lchannel(reader, writer_remote, m(roption.direct), roption.connection_change))
@@ -95,7 +101,7 @@ async def stream_handler(reader, writer, unix, lbind, protos, rserver, cipher, s
         if debug:
             raise
 
-async def reuse_stream_handler(reader, writer, unix, lbind, protos, rserver, urserver, block, cipher, salgorithm, verbose=DUMMY, modstat=lambda r,h:lambda i:DUMMY, **kwargs):
+async def reuse_stream_handler(reader, writer, unix, lbind, protos, rserver, urserver, block, cipher, salgorithm, verbose=DUMMY, modstat=lambda u,r,h:lambda i:DUMMY, **kwargs):
     try:
         if unix:
             remote_ip, server_ip, remote_text = 'local', None, 'unix_local'
@@ -123,7 +129,7 @@ async def reuse_stream_handler(reader, writer, unix, lbind, protos, rserver, urs
             except Exception:
                 writer_remote.close()
                 raise Exception('Unknown remote protocol')
-            m = modstat(remote_ip, host_name)
+            m = modstat(True, remote_ip, host_name)
             asyncio.ensure_future(lproto.channel(reader_remote, writer, m(2+roption.direct), m(4+roption.direct)))
             asyncio.ensure_future(lproto.channel(reader, writer_remote, m(roption.direct), roption.connection_change))
         except Exception as ex:
@@ -149,7 +155,7 @@ async def datagram_handler(writer, data, addr, protos, urserver, block, cipher,
         remote_ip, remote_port, *_ = addr
         remote_text = f'{remote_ip}:{remote_port}'
         data = cipher.datagram.decrypt(data) if cipher else data
-        lproto, host_name, port, data = proto.udp_parse(protos, data, sock=writer.get_extra_info('socket'), **kwargs)
+        lproto, user, host_name, port, data = proto.udp_accept(protos, data, sock=writer.get_extra_info('socket'), **kwargs)
         if host_name == 'echo':
             writer.sendto(data, addr)
         elif host_name == 'empty':
@@ -161,7 +167,7 @@ async def datagram_handler(writer, data, addr, protos, urserver, block, cipher,
             verbose(f'UDP {lproto.name} {remote_text}{roption.logtext(host_name, port)}')
             data = roption.prepare_udp_connection(host_name, port, data)
             def reply(rdata):
-                rdata = lproto.udp_client2(host_name, port, rdata)
+                rdata = lproto.udp_pack(host_name, port, rdata)
                 writer.sendto(cipher.datagram.encrypt(rdata) if cipher else rdata, addr)
             await roption.open_udp_connection(host_name, port, data, addr, reply)
     except Exception as ex:
@@ -230,7 +236,7 @@ class BackwardConnection(object):
                 except asyncio.TimeoutError:
                     data = None
                 if data and data[0] != 0:
-                    reader._buffer[0:0] = data
+                    reader.rollback(data)
                     asyncio.ensure_future(handler(reader, writer))
                 else:
                     writer.close()
@@ -295,7 +301,7 @@ class ProxyURI(object):
                 prot.update = time.perf_counter()
             def datagram_received(prot, data, addr):
                 data = self.cipher.datagram.decrypt(data) if self.cipher else data
-                data = self.rproto.udp_client(data) if not self.direct else data
+                data = self.rproto.udp_unpack(data) if not self.direct else data
                 reply(data)
                 prot.update = time.perf_counter()
             def connection_lost(prot, exc):
@@ -480,7 +486,14 @@ class ProxyURI(object):
             port = int(port) if port else (22 if 'ssh' in rawprotos else 8080)
         else:
             host_name = port = None
-        return ProxyURI(protos=protos, rproto=protos[0], cipher=cipher, auth=url.fragment.encode(), \
+        if url.fragment.startswith('#'):
+            with open(url.fragment[1:]) as f:
+                auth = f.read().rstrip().encode()
+        else:
+            auth = url.fragment.encode()
+        users = [i.rstrip() for i in auth.split(b'\n')] if auth else None
+        auth = users[0] if users else b''
+        return ProxyURI(protos=protos, rproto=protos[0], cipher=cipher, auth=auth, users=users, \
                         match=match, bind=loc or urlpath, host_name=host_name, port=port, \
                         unix=not loc, lbind=lbind, sslclient=sslclient, sslserver=sslserver, \
                         alive=True, direct='direct' in protonames, tunnel='tunnel' in protonames, \
@@ -622,7 +635,7 @@ def main():
             print('exit')
         if args.sys:
             args.sys.clear()
-    for task in asyncio.Task.all_tasks():
+    for task in asyncio.all_tasks(loop) if hasattr(asyncio, 'all_tasks') else asyncio.Task.all_tasks():
         task.cancel()
     for server in servers:
         server.close()
